# coding: utf-8

import json

from django.conf.urls import url
from django.contrib import admin
from django.core.cache import cache
from django.http import HttpResponse, HttpResponseRedirect
from django.utils.translation import ugettext_lazy as _

from django_clear_memcache.clear import ClearMemcacheController, ClearMemcacheNoCacheFoundError
from django_clear_memcache.models import ClearMemcache


try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


########################################################################
class ClearMemcacheAdmin(admin.ModelAdmin):
    model = ClearMemcache
    change_list_template = 'admin/clear_memcache.html'

    # ----------------------------------------------------------------------
    def get_queryset(self, request):
        """Override queryset to return a simple empty queryset
           as we do not really have a real model anyway"""
        return ClearMemcache.objects.none()

    queryset = get_queryset

    # ----------------------------------------------------------------------
    def has_add_permission(self, request):
        """A fake model should not be added"""
        return False

    # ----------------------------------------------------------------------
    def has_delete_permission(self, request, obj=None):
        """A fake model should not be added"""
        return False

    # ----------------------------------------------------------------------
    def get_urls(self):
        urls = admin.ModelAdmin.get_urls(self)
        my_urls = [
            url(r'^clear/$', self.admin_site.admin_view(self.clear), name='clear_cache'),
            url(r'^list/$', self.admin_site.admin_view(self.list_cache_items), name='cache_list_keys'),
        ]
        return my_urls + urls

    # ----------------------------------------------------------------------
    def clear(self, request):
        # clear the cache
        use_prefix = 'clear_prefix' in request.POST
        clear_cache_controller = ClearMemcacheController()
        clear_cache_controller.clear_cache(use_prefix=use_prefix)
        self.message_user(request, _('Cache has been cleared'), fail_silently=True)
        # redirect back to the changelist page
        changelist_url = self._admin_url('changelist')
        return HttpResponseRedirect(changelist_url)

    # ----------------------------------------------------------------------
    def list_cache_items(self, request):
        clear_cache_controller = ClearMemcacheController()
        cache_keys_prefix = clear_cache_controller.keys(use_prefix=True)
        # convert to json and return
        json_ = json.dumps(cache_keys_prefix)
        return HttpResponse(json_, content_type='application/json')

    # ----------------------------------------------------------------------
    def _admin_url(self, target_url):
        opts = self.model._meta
        url_ = "admin:%s_%s_%s" % (opts.app_label, opts.object_name.lower(), target_url)
        return reverse(url_)

    # ----------------------------------------------------------------------
    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or dict()
        extra_context['title'] = _('Clear Memcache')
        try:
            # query the cache
            clear_cache_controller = ClearMemcacheController()
            cache_keys_prefix = clear_cache_controller.keys(use_prefix=True)
            cache_keys = clear_cache_controller.keys(use_prefix=False)
        except ClearMemcacheNoCacheFoundError:
            # ok, we've got no cache or at least no Memcache :(
            extra_context['no_cache_found'] = True
        else:
            extra_context['cache_keys_prefix_count'] = len(cache_keys_prefix)
            extra_context['cache_key_count'] = len(cache_keys)
            extra_context['cache_key_prefix'] = cache.key_prefix
            extra_context['cl'] = cache_keys
        # create the response
        response = admin.ModelAdmin.changelist_view(
            self,
            request,
            extra_context=extra_context)

        return response


admin.site.register(ClearMemcache, ClearMemcacheAdmin)



# -*- coding: utf-8 -*-

# MemcachedUtility is loosely based on https://github.com/dlrust/python-memcached-stats

from django.conf import settings
from django.core.cache import cache, DEFAULT_CACHE_ALIAS
from django.core.exceptions import ImproperlyConfigured
from django.utils import six

from django_clear_memcache.utility import MemcachedUtility


########################################################################
class ClearMemcacheNoCacheFoundError(ImproperlyConfigured):
    pass


########################################################################
class ClearMemcacheController(object):

    # ----------------------------------------------------------------------
    def __init__(self):
        self._servers = None
        self._host = None
        self._port = None
        self._cache_key_prefix = None
        self._keys = None
        self._init()

    # ----------------------------------------------------------------------
    def _init(self):
        # support django-debug-toolbar which wraps the cache instance
        cache_ = getattr(cache, 'cache', cache)
        # check if we got a usable cache instance
        if not hasattr(cache_, '_lib'):
            raise ClearMemcacheNoCacheFoundError(
                'Unknown memcached backend or no memcached backend found')

        # get cache key prefix
        self._cache_key_prefix = cache.key_prefix
        # LOCATION can be a string or a sequence
        server = settings.CACHES[DEFAULT_CACHE_ALIAS]['LOCATION']
        if isinstance(server, six.string_types):
            servers = server.split(';')
        else:
            servers = server

        self._servers = list()
        for server in servers:
            host, port = server.split(':')
            port = self._parse_port(port)
            if self._host == 'unix':
                # for now, ignore unix domain sockets
                continue
            self._servers.append((host, port))

    # ----------------------------------------------------------------------
    def _parse_port(self, port):
        try:
            return int(port)
        except (ValueError, TypeError) as e:
            raise ClearMemcacheNoCacheFoundError('Unable to parse port "{}": {}'.format(port, e))

    # ----------------------------------------------------------------------
    def keys(self, use_prefix=True, refresh=False):
        # cache keys per instance
        if self._keys is None or refresh:
            self._keys = list()
            for self._host, self._port in self._servers:
                try:
                    utility = MemcachedUtility(self._host, self._port)
                    utility.open()
                    # search for keys, according to the configured prefix
                    server_keys = list(utility.keys())
                    self._keys.extend(server_keys)
                finally:
                    utility.close()

        if use_prefix:
            keys = self._keys[:]  # copy
            for key in self._keys:
                if not key.startswith(self._cache_key_prefix):
                    keys.remove(key)
        else:
            keys = self._keys

        return keys

    # ----------------------------------------------------------------------
    def clear_cache(self, use_prefix=True):
        if use_prefix:
            # iterate over all caches and delete single keys
            for self._host, self._port in self._servers:
                # clear this cache
                self._clear()
        else:
            # flush all caches
            for self._host, self._port in self._servers:
                # clear this cache
                self._flush()

    # ----------------------------------------------------------------------
    def _clear(self):
        utility = MemcachedUtility(self._host, self._port)
        utility.open()
        try:
            for key in list(utility.keys()):
                key = key.strip()
                if key.startswith(self._cache_key_prefix):
                    utility.delete(key)
        finally:
            utility.close()

    # ----------------------------------------------------------------------
    def _flush(self):
        utility = MemcachedUtility(self._host, self._port)
        utility.open()
        try:
            utility.flush()
            # now iterate over all remaining keys and retrieve them to finally clear them from
            # the cache, otherwise they would still show up even if already flushed
            for key in list(utility.keys()):
                utility.get(key)
        finally:
            utility.close()



# -*- coding: utf-8 -*-

try:
    # try to use AppConfig from Django 1.7, silently ignore on older versions
    from django.apps import AppConfig
    from django.utils.translation import ugettext_lazy as _
except ImportError:
    pass
else:
    class ClearMemcacheConfig(AppConfig):
        name = 'django_clear_memcache'
        verbose_name = _("Clear Memcache")



# -*- coding: utf-8 -*-

from django.db import models
from django.utils.translation import ugettext_lazy as _


########################################################################
class ClearMemcache(models.Model):
    """This is a fake model, just to trick Django's admin to
       have an easy changelist view"""

    class Meta:
        managed = False
        app_label = 'django_clear_memcache'
        verbose_name = _('Clear Memcache')
        verbose_name_plural = _('Clear Memcache')



# -*- coding: utf-8 -*-

import sys
from shutil import rmtree
from setuptools import setup, find_packages

NAME = 'django-clear-memcache'
VERSION = '1.4.0'

if 'bdist_wheel' in sys.argv:
    # Remove previous build dir when creating a wheel build, since if files have been removed
    # from the project, they'll still be cached in the build dir and end up as part of the
    # build, which is really neat!
    for directory in ('build', 'dist', 'django_clear_memcache.egg-info'):
        try:
            rmtree(directory)
        except Exception:
            pass


setup(
    name=NAME,
    description='Allow to easily clear Memcache items in Django\'s admin interface.',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    version=VERSION,
    author='Enrico Tröger',
    author_email='enrico.troeger@uvena.de',
    url='https://github.com/eht16/django-clear-memcache',
    download_url='https://github.com/eht16/django-clear-memcache/archive/django-clear-memcache-%s.tar.gz' % VERSION,
    packages=find_packages(),
    license="BSD",
    include_package_data=True,
    zip_safe=False,
    requires=['django'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Framework :: Django',
    ],
)



# -*- coding: utf-8 -*-

# MemcachedUtility is loosely based on https://github.com/dlrust/python-memcached-stats

from logging import getLogger
import re
import telnetlib


TELNET_READ_TIMEOUT = 1.0


########################################################################
class MemcachedUtility(object):
    """
    A small utility to control a memcached server via its telnet interface.
    """

    _key_regex = re.compile(r'ITEM (.*) \[(.*); (.*)\]')
    _slab_regex = re.compile(r'STAT items:(.*):number')
    _stat_regex = re.compile(r"STAT (.*) (.*)\r")

    # ----------------------------------------------------------------------
    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._client = None
        self._logger = getLogger('djago_memcache_clear')

    # ----------------------------------------------------------------------
    def open(self):
        if self._client is None:
            self._client = telnetlib.Telnet(self._host, self._port)

    # ----------------------------------------------------------------------
    def close(self):
        if self._client is not None:
            self._client.close()

    # ----------------------------------------------------------------------
    def command(self, cmd, read_until='END'):
        """Write a command to telnet and return the response"""
        self._logger.debug(
            '[%12s] executing memcache command(%-10s): %s',
            self._host, repr(read_until), cmd)
        command = '{}\n'.format(cmd)
        self._client.write(command.encode())
        read_until_bytes = read_until.encode()
        response = self._client.read_until(read_until_bytes, TELNET_READ_TIMEOUT)
        return response.decode()

    # ----------------------------------------------------------------------
    def key_details(self, sort=True, limit=100):
        """Return a list of tuples containing keys and details"""
        cmd = 'stats cachedump {} {}'
        keys = [key for id_ in self.slab_ids()
                for key in self._key_regex.findall(self.command(cmd.format(id_, limit)))]
        if sort:
            return sorted(keys)
        else:
            return keys

    # ----------------------------------------------------------------------
    def get(self, key):
        """Retrieve a single key"""
        return self.command('get {}'.format(key))

    # ----------------------------------------------------------------------
    def flush(self):
        """Retrieve a single key"""
        return self.command('flush_all', read_until='OK')

    # ----------------------------------------------------------------------
    def keys(self, sort=True, limit=100):
        """Return a list of keys in use"""
        return [key[0] for key in self.key_details(sort=sort, limit=limit)]

    # ----------------------------------------------------------------------
    def slab_ids(self):
        """Return a list of slab ids in use"""
        return self._slab_regex.findall(self.command('stats items'))

    # ----------------------------------------------------------------------
    def stats(self):
        """Return a dict containing memcached stats"""
        return dict(self._stat_regex.findall(self.command('stats')))

    # ----------------------------------------------------------------------
    def delete(self, key):
        result = self.command('delete {}'.format(key), read_until='\n')
        result = result.strip()
        self._logger.debug('Delete for key %s returned: %s', key, result)
